# __init__.py

# Only expose the submodules, not individual classes
from . import core
from . import constraints
from . import datasets
from . import descriptor
from . import learners
from . import metrics
from . import networks

# Define __all__ to specify that the submodules are accessible, but not classes directly.
__all__ = [
    "core",
    "constraints",
    "datasets",
    "descriptor",
    "learners",
    "metrics",
    "networks"
]


from torch.nn import Linear, Sequential, ReLU, Module


class MLPNetwork(Module):
    """
    A multi-layer perceptron (MLP) neural network model consisting of
    an input layer, multiple hidden layers, and an output layer.

    This class constructs an MLP with configurable hyperparameters such as the
    number of input features, output features, number of hidden layers, and
    the dimensionality of hidden layers. It provides methods for both
    building the model and performing a forward pass through the network.

    Attributes:
        n_inputs (int): The number of input features.
        n_outputs (int): The number of output features.
        n_hidden_layers (int): The number of hidden layers in the network.
        hidden_dim (int): The dimensionality of the hidden layers.
        input (nn.Module): The input layer (linear transformation followed by ReLU).
        hidden (nn.Module): The sequential hidden layers (each consisting of
                             a linear transformation followed by ReLU).
        out (nn.Module): The output layer (linear transformation).
    """

    def __init__(
        self,
        n_inputs=25,
        n_outputs=2,
        n_hidden_layers=2,
        hidden_dim=35,
    ):
        """
        Initializes the MLP network with the given hyperparameters.

        Args:
            n_inputs (int, optional): The number of input features. Defaults to 25.
            n_outputs (int, optional): The number of output features. Defaults to 2.
            n_hidden_layers (int, optional): The number of hidden layers. Defaults to 2.
            hidden_dim (int, optional): The dimensionality of the hidden layers. Defaults to 35.
        """
        super().__init__()

        # Init object variables
        self.n_inputs = n_inputs
        self.n_outputs = n_outputs
        self.n_hidden_layers = n_hidden_layers
        self.hidden_dim = hidden_dim

        # Set up the components of our model
        self.input = self.linear(self.n_inputs, self.hidden_dim)
        self.hidden = Sequential(
            *(
                self.linear(self.hidden_dim, self.hidden_dim)
                for _ in range(n_hidden_layers)
            )
        )
        self.out = Linear(self.hidden_dim, self.n_outputs)

    def forward(self, X):
        """
        Performs a forward pass through the network.

        Args:
            X (Tensor): The input tensor to be passed through the network.

        Returns:
            dict: A dictionary containing the 'input' (original input) and
                  'output' (predicted output) of the network.
        """
        input = X
        output = self.out(self.hidden(self.input(X)))

        return {"input": input, "output": output}

    @staticmethod
    def linear(in_features, out_features):
        """
        Creates a basic linear block with a linear transformation followed
        by a ReLU activation function.

        Args:
            in_features (int): The number of input features.
            out_features (int): The number of output features.

        Returns:
            nn.Module: A sequential module consisting of a Linear layer and ReLU activation.
        """
        return Sequential(
            Linear(in_features, out_features),
            ReLU(),
        )



import logging
from typing import Dict
from lightning import LightningModule
from torch import Tensor, float32, no_grad, norm, tensor
from torchmetrics import Metric
from torch.nn import ModuleDict

from .constraints import Constraint
from .metrics import ConstraintSatisfactionRatio
from .descriptor import Descriptor


class CGGDModule(LightningModule):
    """
    A PyTorch Lightning module that integrates constraint-guided optimization into the training and validation steps.

    This module extends the `LightningModule` and incorporates constraints on the neural network's predictions 
    by adjusting the loss using a rescale factor. The constraints are checked, and the loss is modified to guide 
    the optimization process based on these constraints.

    Attributes:
        descriptor (Descriptor): The object that describes the layers and neurons of the network, including 
                                 the categorization of variable layers.
        constraints (list[Constraint]): A list of constraints that define the conditions to guide the optimization.
        train_csr (Dict[str, Metric]): A dictionary of `ConstraintSatisfactionRatio` metrics to track constraint satisfaction 
                                       during training, indexed by constraint name.
        valid_csr (Dict[str, Metric]): A dictionary of `ConstraintSatisfactionRatio` metrics to track constraint satisfaction 
                                       during validation, indexed by constraint name.
    """

    def __init__(self, descriptor: Descriptor, constraints: list[Constraint]):
        """
        Initializes the CGGDModule with a descriptor and a list of constraints.

        Args:
            descriptor (Descriptor): The object that describes the network's layers and neurons, including their categorization.
            constraints (list[Constraint]): A list of constraints that will guide the optimization process.

        Raises:
            Warning if there are no variable layers in the descriptor, as constraints will not be applied.
        """

        # Init parent class
        super().__init__()

        # Init object variables
        self.descriptor = descriptor
        self.constraints = constraints

        # Perform checks
        if len(self.descriptor.variable_layers) == 0:
            logging.warning(
                "The descriptor object has no variable layers. The constraint guided loss adjustment is therefore not used. Is this the intended behaviour?"
            )

        # Assign descriptor to constraints
        for constraint in self.constraints:
            constraint.descriptor = descriptor
            constraint.run_init_descriptor()

        # Init constraint metric logging
        self.train_csr: Dict[str, Metric] = ModuleDict(
            {
                constraint.constraint_name: ConstraintSatisfactionRatio()
                for constraint in self.constraints
            }
        )
        self.train_csr["global"] = ConstraintSatisfactionRatio()
        self.valid_csr: Dict[str, Metric] = ModuleDict(
            {
                constraint.constraint_name: ConstraintSatisfactionRatio()
                for constraint in self.constraints
            }
        )
        self.valid_csr["global"] = ConstraintSatisfactionRatio()

    def training_step(
        self,
        prediction: dict[str, Tensor],
        loss: Tensor,
    ):
        """
        The training step where the standard loss is combined with rescale loss based on the constraints.

        For each constraint, the satisfaction ratio is checked, and the loss is adjusted by adding a rescale loss
        based on the directions calculated by the constraint.

        Args:
            prediction (dict[str, Tensor]): The model's predictions for each layer.
            loss (Tensor): The base loss from the model's forward pass.

        Returns:
            Tensor: The combined loss, including both the original loss and the rescale loss from the constraints.
        """

        # Init scalar tensor for loss
        total_rescale_loss = tensor(0, dtype=float32, device=self.device)

        # Compute rescale loss without tracking gradients
        with no_grad():

            # For each constraint, TODO split into real and validation only constraints
            for constraint in self.constraints:

                # Check if constraints are satisfied and calculate directions
                constraint_checks = constraint.check_constraint(prediction)
                constraint_directions = constraint.calculate_direction(prediction)

                # Only do direction calculations for variable layers affecting constraint
                for layer in constraint.layers & self.descriptor.variable_layers:

                    # Multiply direction modifiers with constraint result
                    constraint_result = (
                        constraint_checks[layer].unsqueeze(1).type(float32)
                        * constraint_directions[layer]
                    )

                    # Multiply result with rescale factor o constraint
                    constraint_result *= constraint.rescale_factor

                    # Calculate gradients of general loss for each sample
                    loss.backward(retain_graph=True, inputs=prediction[layer])
                    loss_grad = prediction[layer].grad

                    # Calculate loss gradient norm
                    norm_loss_grad = norm(loss_grad, dim=0, p=2, keepdim=True)

                    # Calculate rescale loss
                    rescale_loss = (
                        (prediction[layer] * constraint_result * norm_loss_grad)
                        .sum()
                        .abs()
                    )

                    # Store rescale loss for this reference space
                    total_rescale_loss += rescale_loss

                    # Log constraint satisfaction ratio
                    # NOTE does this take into account spaces with different dimensions?
                    self.train_csr[constraint.constraint_name](constraint_checks[layer])
                    self.train_csr["global"](constraint_checks[layer])
                    self.log(
                        f"train_csr_{constraint.constraint_name}_{layer}",
                        self.train_csr[constraint.constraint_name],
                        on_step=False,
                        on_epoch=True,
                    )

        # Log global constraint satisfaction ratio
        self.log(
            "train_csr_global",
            self.train_csr["global"],
            on_step=False,
            on_epoch=True,
        )

        # Return combined loss
        return loss + total_rescale_loss

    def validation_step(
        self,
        prediction: dict[str, Tensor],
        loss: Tensor,
    ):
        """
        The validation step where the satisfaction of constraints is checked without applying the rescale loss.

        Similar to the training step, but without updating the loss, this method tracks the constraint satisfaction 
        during validation.

        Args:
            prediction (dict[str, Tensor]): The model's predictions for each layer.
            loss (Tensor): The base loss from the model's forward pass.

        Returns:
            Tensor: The base loss value for validation.
        """

        # Compute rescale loss without tracking gradients
        with no_grad():

            # For each constraint in this reference space, calculate directions
            for constraint in self.constraints:

                # Check if constraints are satisfied for
                constraint_checks = constraint.check_constraint(prediction)

                # Only do direction calculations for variable layers affecting constraint
                for layer in constraint.layers & self.descriptor.variable_layers:

                    # Log constraint satisfaction ratio
                    # NOTE does this take into account spaces with different dimensions?
                    self.valid_csr[constraint.constraint_name](constraint_checks[layer])
                    self.valid_csr["global"](constraint_checks[layer])
                    self.log(
                        f"valid_csr_{constraint.constraint_name}",
                        self.valid_csr[constraint.constraint_name],
                        on_step=False,
                        on_epoch=True,
                    )

        # Log global constraint satisfaction ratio
        self.log(
            "valid_csr_global",
            self.valid_csr["global"],
            on_step=False,
            on_epoch=True,
        )

        # Return loss
        return loss



class Descriptor:
    """
    A class to manage the mapping of neurons to layers and their properties
    (e.g., output, constant, or variable) in a neural network.

    This class enables the organization and description of network elements,
    such as associating neurons with specific layers and categorizing layers
    as outputs, constants, or variables.

    This allows users to easily place constraints on parts of the network by
    referencing the name that is configured in this class.
    """

    def __init__(
        self,
    ):
        """
        Initialize the Descriptor class with empty mappings for neurons and layers.

        This includes:
            - `neuron_to_layer`: A dictionary mapping neuron names to their corresponding layer names.
            - `neuron_to_index`: A dictionary mapping neuron names to their corresponding index within a layer.
            - `output_layers`: A set that holds the names of layers marked as output layers.
            - `constant_layers`: A set that holds the names of layers marked as constant layers.
            - `variable_layers`: A set that holds the names of layers marked as variable layers.
        """

        # Define dictionaries that will translate neuron names to layer and index
        self.neuron_to_layer: dict[str, str] = {}
        self.neuron_to_index: dict[str, int] = {}

        # Define sets that will hold the layers based on which type
        self.output_layers: set[str] = set()
        self.constant_layers: set[str] = set()
        self.variable_layers: set[str] = set()

    def add(
        self,
        layer_name: str,
        neuron_names: list[str],
        output: bool = False,
        constant: bool = False,
    ):
        """
        Add a layer to the descriptor, associating it with neurons and marking it
        as an output or constant layer.

        Args:
            layer_name (str): The name of the layer to be added.
            neuron_names (list[str]): A list of neuron names that belong to the layer.
            output (bool, optional): If True, mark this layer as an output layer. Defaults to False.
            constant (bool, optional): If True, mark this layer as a constant layer. Defaults to False.
        """

        if output:
            self.output_layers.add(layer_name)

        if constant:
            self.constant_layers.add(layer_name)
        else:
            self.variable_layers.add(layer_name)

        for index, neuron_name in enumerate(neuron_names):
            self.neuron_to_layer[neuron_name] = layer_name
            self.neuron_to_index[neuron_name] = index



import os
from urllib.error import URLError
import numpy as np
from pathlib import Path
from typing import Callable, Optional, Union
import pandas as pd
import lightning as L
from torch.utils.data import Dataset, random_split, DataLoader
import torch

from torchvision.datasets.utils import check_integrity, download_and_extract_archive


class BiasCorrection(Dataset):
    """
    Bias Correction Dataset.

    This class provides access to a dataset used for bias correction of numerical prediction models,
    specifically for temperature forecasts. The dataset is stored in CSV format and can be optionally
    transformed using a provided function. The class includes methods for downloading, loading, and
    accessing the dataset, as well as applying transformations.

    Args:
        root (str or pathlib.Path): Root directory of the dataset where the 'BiasCorrection/Bias_correction_ucl.csv' file is located.
        download (bool, optional): If True, the dataset will be downloaded from the internet and saved to the root directory. 
                                   If the dataset already exists, it will not be downloaded again. Default is False.
        transform (callable, optional): A function/transform that takes a Pandas DataFrame and returns a transformed version.
                                        This allows for data preprocessing before it is loaded. Default is None.

    Acknowledgement:
        This class was developed with inspiration from the MNIST dataset class in torchvision.
    """

    mirrors = [
        "https://archive.ics.uci.edu/static/public/514/",
    ]

    resources = [
        (
            "bias+correction+of+numerical+prediction+model+temperature+forecast.zip",
            "3deee56d461a2686887c4ae38fe3ccf3",
        ),
    ]

    def __init__(
        self,
        root: Union[str, Path],
        download: bool = False,
        transform: Optional[Callable] = None,
    ) -> None:
        """
        Initializes the BiasCorrection dataset, optionally downloading it if necessary.

        Args:
            root (Union[str, Path]): The root directory where the dataset will be stored or accessed.
            download (bool, optional): If True, downloads the dataset from the provided mirrors. Default is False.
            transform (Optional[Callable], optional): A function to transform the data before loading. Default is None.
        """

        super().__init__()
        self.root = root
        self.transform = transform

        if download:
            self.download()

        if not self._check_exists():
            raise RuntimeError(
                "Dataset not found. You can use download=True to download it"
            )

        self.data_input, self.data_output = self._load_data()

    def _load_data(self):
        """
        Loads the dataset from the CSV file, applies transformations (if provided), and separates the data into input and output variables.

        Returns:
            Tuple: A tuple containing two numpy arrays: `data_input` (input data) and `data_output` (output data).
        """

        if self.transform:
            data: pd.DataFrame = (
                pd.read_csv(os.path.join(self.data_folder, "Bias_correction_ucl.csv"))
                .pipe(self.transform)
                .pipe(self.add_input_output_temperature)
            )
        else:
            data: pd.DataFrame = pd.read_csv(
                os.path.join(self.data_folder, "Bias_correction_ucl.csv")
            ).pipe(self.add_input_output_temperature)

        data_input = data["Input"].to_numpy(dtype=np.float32)
        data_output = data["Output"].to_numpy(dtype=np.float32)

        return data_input, data_output

    def __len__(self):
        """
        Returns the number of examples in the dataset.

        Returns:
            int: The total number of examples in the dataset (i.e., the number of rows in the input data).
        """
        
        return self.data_input.shape[0]

    def __getitem__(self, idx):
        """
        Retrieves a single example and its corresponding target from the dataset.

        Args:
            idx (int): The index of the example to retrieve.

        Returns:
            Tuple: A tuple containing two tensors: the input example and the target output.
        """

        example = self.data_input[idx, :]
        target = self.data_output[idx, :]
        example = torch.tensor(example)
        target = torch.tensor(target)
        return example, target

    @property
    def data_folder(self) -> str:
        """
        Returns the path to the folder where the dataset is stored.

        Returns:
            str: The path to the dataset folder within the root directory.
        """

        return os.path.join(self.root, self.__class__.__name__)

    def _check_exists(self) -> bool:
        """
        Checks if the dataset files exist and their integrity is verified.

        Returns:
            bool: True if the dataset exists and is valid, otherwise False.
        """

        return all(
            check_integrity(os.path.join(self.data_folder, file_path), checksum)
            for file_path, checksum in self.resources
        )

    def download(self) -> None:
        """
        Downloads the dataset from the provided mirrors and extracts it.

        The method ensures the dataset is downloaded and extracted to the appropriate folder. 
        If the dataset is already present, it will not be downloaded again.
        """

        if self._check_exists():
            return

        os.makedirs(self.data_folder, exist_ok=True)

        # download files
        for filename, md5 in self.resources:
            errors = []
            for mirror in self.mirrors:
                url = f"{mirror}{filename}"
                try:
                    download_and_extract_archive(
                        url, download_root=self.data_folder, filename=filename, md5=md5
                    )
                except URLError as e:
                    errors.append(e)
                    continue
                break
            else:
                s = f"Error downloading {filename}:\n"
                for mirror, err in zip(self.mirrors, errors):
                    s += f"Tried {mirror}, got:\n{str(err)}\n"
                raise RuntimeError(s)

    @staticmethod
    def add_input_output_temperature(df: pd.DataFrame) -> pd.DataFrame:
        """Add a multiindex denoting if the column is an input or output variable."""
        # copy the dataframe
        temp_df = df.copy()
        # extract all the column names
        column_names = temp_df.columns.tolist()
        # only the last 2 columns are output variables, all others are input variables. So make list of corresponding lengths of 'Input' and 'Output'
        input_list = ["Input"] * (len(column_names) - 2)
        output_list = ["Output"] * 2
        # concat both lists
        input_output_list = input_list + output_list
        # define multi index for attaching this 'Input' and 'Output' list with the column names already existing
        multiindex_bias = pd.MultiIndex.from_arrays([input_output_list, column_names])
        # transpose such that index can be adjusted to multi index
        new_df = pd.DataFrame(df.transpose().to_numpy(), index=multiindex_bias)
        # transpose back such that columns are the same as before except with different labels
        return new_df.transpose()


class BCDataModule(L.LightningDataModule):
    """Bias Correction dataset module."""

    def __init__(
        self,
        dataset_directory: str = "./datasets",
        batch_size: int = 32,
        train_size: float = 0.8,
        val_size: float = 0.1,
        test_size: float = 0.1,
        shuffle_train: bool = True,
        shuffle_val: bool = False,
        shuffle_test: bool = False,
        num_workers: int = 4,
        pin_memory: bool = False,
    ) -> None:
        super().__init__()
        # Define required parameters here
        self.batch_size = batch_size
        self.dataset_directory = dataset_directory
        self.train_size = train_size
        self.val_size = val_size
        self.test_size = test_size
        self.shuffle_train = shuffle_train
        self.shuffle_val = shuffle_val
        self.shuffle_test = shuffle_test
        self.num_workers = num_workers
        self.pin_memory = pin_memory

    def prepare_data(self):
        # Define steps that should be done
        # on only one GPU, like getting data.
        BiasCorrection(self.dataset_directory, download=True, transform=self.transform)

    def setup(self, stage=None):
        # Define steps that should be done on
        # every GPU, like splitting data, applying
        # transform etc.
        data = BiasCorrection(self.dataset_directory, transform=self.transform)

        train_val_data, self.test_data = random_split(
            data, [1 - self.test_size, self.test_size]
        )
        self.train_data, self.val_data = random_split(
            train_val_data,
            [
                self.train_size / (1 - self.test_size),
                self.val_size / (1 - self.test_size),
            ],
        )

    def train_dataloader(self):
        # Return DataLoader for Training Data here
        return DataLoader(
            self.train_data,
            batch_size=self.batch_size,
            shuffle=self.shuffle_train,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

    def val_dataloader(self):
        # Return DataLoader for Validation Data here
        return DataLoader(
            self.val_data,
            batch_size=self.batch_size,
            shuffle=self.shuffle_val,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

    def test_dataloader(self):
        # Return DataLoader for Testing Data here
        return DataLoader(
            self.test_data,
            batch_size=self.batch_size,
            shuffle=self.shuffle_test,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

    @staticmethod
    def transform(df: pd.DataFrame) -> pd.DataFrame:
        def date_to_datetime(df: pd.DataFrame) -> pd.DataFrame:
            """Transform the string that denotes the date to the datetime format in pandas."""
            # make copy of dataframe
            df_temp = df.copy()
            # add new column at the front where the date string is transformed to the datetime format
            df_temp.insert(0, "DateTransformed", pd.to_datetime(df_temp["Date"]))
            return df_temp

        def add_year(df: pd.DataFrame) -> pd.DataFrame:
            """Extract the year from the datetime cell and add it as a new column to the dataframe at the front."""
            # make copy of dataframe
            df_temp = df.copy()
            # extract year and add new column at the front containing these numbers
            df_temp.insert(0, "Year", df_temp["DateTransformed"].dt.year)
            return df_temp

        def add_month(df: pd.DataFrame) -> pd.DataFrame:
            """Extract the month from the datetime cell and add it as a new column to the dataframe at the front."""
            # make copy of dataframe
            df_temp = df.copy()
            # extract month and add new column at index 1 containing these numbers
            df_temp.insert(1, "Month", df_temp["DateTransformed"].dt.month)
            return df_temp

        def add_day(df: pd.DataFrame) -> pd.DataFrame:
            """Extract the day from the datetime cell and add it as a new column to the dataframe at the front."""
            # make copy of dataframe
            df_temp = df.copy()
            # extract day and add new column at index 2 containing these numbers
            df_temp.insert(2, "Day", df_temp["DateTransformed"].dt.day)
            return df_temp

        def normalize_columns_bias(df: pd.DataFrame) -> pd.DataFrame:
            """Normalize the columns for the bias correction dataset. This is different from normalizing all the columns separately because the
            upper and lower bounds for the output variables are assumed to be the same.
            """
            # copy the dataframe
            temp_df = df.copy()
            # normalize each column
            for feature_name in df.columns:
                # the output columns are normalized using the same upper and lower bound for more efficient check of the inequality
                if feature_name == "Next_Tmax" or feature_name == "Next_Tmin":
                    max_value = 38.9
                    min_value = 11.3
                # the input columns are normalized using their respective upper and lower bounds
                else:
                    max_value = df[feature_name].max()
                    min_value = df[feature_name].min()
                temp_df[feature_name] = (df[feature_name] - min_value) / (
                    max_value - min_value
                )
            return temp_df

        def sample_2500_examples(df: pd.DataFrame) -> pd.DataFrame:
            """Sample 2500 examples from the dataframe without replacement."""
            temp_df = df.copy()
            sample_df = temp_df.sample(n=2500, replace=False, random_state=3, axis=0)
            return sample_df

        return (
            # drop missing values
            df.dropna(how="any")
            # transform string date to datetime format
            .pipe(date_to_datetime)
            # add year as a single column
            .pipe(add_year)
            # add month as a single column
            .pipe(add_month)
            # add day as a single column
            .pipe(add_day)
            # remove original date string and the datetime format
            .drop(["Date", "DateTransformed"], axis=1, inplace=False)
            # convert all numbers to float32
            .astype("float32")
            # normalize columns
            .pipe(normalize_columns_bias)
            # sample 2500 examples out of the dataset
            .pipe(sample_2500_examples)
        )


class FiniteIncome(Dataset):
    """Finite Income Dataset.

    Args:
        root (str or ``pathlib.Path``): Root directory of dataset where ``BiasCorrection/Bias_correction_ucl.csv`` exists.
        download (bool, optional): If True, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        transform (callable, optional): A function/transform that takes in a Pandas DataFrame
            and returns a transformed version.

    Acknowledgement:
        This class was developed with inspiration from the MNIST dataset class in torchvision.
    """

    mirrors = [
        "https://www.kaggle.com/api/v1/datasets/download/grosvenpaul/",
    ]

    resources = [
        (
            "family-income-and-expenditure",
            "7d74bc7facc3d7c07c4df1c1c6ac563e",
        ),
    ]

    def __init__(
        self,
        root: Union[str, Path],
        download: bool = False,
        transform: Optional[Callable] = None,
    ) -> None:
        super().__init__()
        self.root = root
        self.transform = transform

        if download:
            self.download()

        if not self._check_exists():
            raise RuntimeError(
                "Dataset not found. You can use download=True to download it"
            )

        self.data_input, self.data_output = self._load_data()

    def _load_data(self):

        if self.transform:
            data: pd.DataFrame = pd.read_csv(
                os.path.join(self.data_folder, "Family Income and Expenditure.csv")
            ).pipe(self.transform)
        else:
            data: pd.DataFrame = pd.read_csv(
                os.path.join(self.data_folder, "Family Income and Expenditure.csv")
            ).pipe(self.add_input_output_family_income)

        data_input = data["Input"].to_numpy(dtype=np.float32)
        data_output = data["Output"].to_numpy(dtype=np.float32)

        return data_input, data_output

    def __len__(self):
        return self.data_input.shape[0]

    def __getitem__(self, idx):
        example = self.data_input[idx, :]
        target = self.data_output[idx, :]
        example = torch.tensor(example)
        target = torch.tensor(target)
        return example, target

    @property
    def data_folder(self) -> str:
        return os.path.join(self.root, self.__class__.__name__)

    def _check_exists(self) -> bool:
        return all(
            check_integrity(os.path.join(self.data_folder, file_path), checksum)
            for file_path, checksum in self.resources
        )

    def download(self) -> None:
        """Download the MNIST data if it doesn't exist already."""

        if self._check_exists():
            return

        os.makedirs(self.data_folder, exist_ok=True)

        # download files
        for filename, md5 in self.resources:
            errors = []
            for mirror in self.mirrors:
                url = f"{mirror}{filename}"
                try:
                    download_and_extract_archive(
                        url, download_root=self.data_folder, filename=filename, md5=md5
                    )
                except URLError as e:
                    errors.append(e)
                    continue
                break
            else:
                s = f"Error downloading {filename}:\n"
                for mirror, err in zip(self.mirrors, errors):
                    s += f"Tried {mirror}, got:\n{str(err)}\n"
                raise RuntimeError(s)

    @staticmethod
    def add_input_output_family_income(df: pd.DataFrame) -> pd.DataFrame:
        """Add a multiindex denoting if the column is an input or output variable."""
        # copy the dataframe
        temp_df = df.copy()
        # extract all the column names
        column_names = temp_df.columns.tolist()
        # the 2nd-9th columns correspond to output variables and all others to input variables. So make list of corresponding lengths of 'Input' and 'Output'
        input_list_start = ["Input"]
        input_list_end = ["Input"] * (len(column_names) - 9)
        output_list = ["Output"] * 8
        # concat both lists
        input_output_list = input_list_start + output_list + input_list_end
        # define multi index for attaching this 'Input' and 'Output' list with the column names already existing
        multiindex_bias = pd.MultiIndex.from_arrays([input_output_list, column_names])
        # transpose such that index can be adjusted to multi index
        new_df = pd.DataFrame(df.transpose().to_numpy(), index=multiindex_bias)
        # transpose back such that columns are the same as before except with different labels
        return new_df.transpose()


class FIDataModule(L.LightningDataModule):
    """Finite Income dataset module."""

    def __init__(
        self,
        dataset_directory: str = "./datasets",
        batch_size: int = 32,
        train_size: float = 0.8,
        val_size: float = 0.1,
        test_size: float = 0.1,
        shuffle_train: bool = True,
        shuffle_val: bool = False,
        shuffle_test: bool = False,
        num_workers: int = 4,
        pin_memory: bool = False,
    ) -> None:
        super().__init__()
        # Define required parameters here
        self.batch_size = batch_size
        self.dataset_directory = dataset_directory
        self.train_size = train_size
        self.val_size = val_size
        self.test_size = test_size
        self.shuffle_train = shuffle_train
        self.shuffle_val = shuffle_val
        self.shuffle_test = shuffle_test
        self.num_workers = num_workers
        self.pin_memory = pin_memory

    def prepare_data(self):
        # Define steps that should be done
        # on only one GPU, like getting data.
        # TODO downloading currently disabled since not compatible with api
        # FiniteIncome(self.dataset_directory, download=True, transform=self.transform)
        pass

    def setup(self, stage=None):
        # Define steps that should be done on
        # every GPU, like splitting data, applying
        # transform etc.
        data = FiniteIncome(self.dataset_directory, transform=self.transform)

        train_val_data, self.test_data = random_split(
            data, [1 - self.test_size, self.test_size]
        )
        self.train_data, self.val_data = random_split(
            train_val_data,
            [
                self.train_size / (1 - self.test_size),
                self.val_size / (1 - self.test_size),
            ],
        )

    def train_dataloader(self):
        # Return DataLoader for Training Data here
        return DataLoader(
            self.train_data,
            batch_size=self.batch_size,
            shuffle=self.shuffle_train,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

    def val_dataloader(self):
        # Return DataLoader for Validation Data here
        return DataLoader(
            self.val_data,
            batch_size=self.batch_size,
            shuffle=self.shuffle_val,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

    def test_dataloader(self):
        # Return DataLoader for Testing Data here
        return DataLoader(
            self.test_data,
            batch_size=self.batch_size,
            shuffle=self.shuffle_test,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

    @staticmethod
    def transform(df: pd.DataFrame) -> pd.DataFrame:
        def normalize_columns_income(df: pd.DataFrame) -> pd.DataFrame:
            """Normalize the columns for the Family Income dataframe. This can also be applied to other dataframes because this function normalizes
            all columns individually."""
            # copy the dataframe
            temp_df = df.copy()
            # normalize each column
            for feature_name in df.columns:
                max_value = df[feature_name].max()
                min_value = df[feature_name].min()
                temp_df[feature_name] = (df[feature_name] - min_value) / (
                    max_value - min_value
                )
            return temp_df

        def check_constraints_income(df: pd.DataFrame) -> pd.DataFrame:
            """Check if all the constraints are satisfied for the dataframe and remove the examples that do not satisfy the constraint. This
            function only works for the Family Income dataset and the constraints are that the household income is larger than all the expenses
            and the food expense is larger than the sum of the other (more detailed) food expenses.
            """
            temp_df = df.copy()
            # check that household income is larger than expenses in the output
            input_array = temp_df["Input"].to_numpy()
            income_array = np.add(
                np.multiply(
                    input_array[:, [0, 1]],
                    np.subtract(
                        np.asarray([11815988, 9234485]), np.asarray([11285, 0])
                    ),
                ),
                np.asarray([11285, 0]),
            )
            expense_array = temp_df["Output"].to_numpy()
            expense_array = np.add(
                np.multiply(
                    expense_array,
                    np.subtract(
                        np.asarray(
                            [
                                791848,
                                437467,
                                140992,
                                74800,
                                2188560,
                                1049275,
                                149940,
                                731000,
                            ]
                        ),
                        np.asarray([3704, 0, 0, 0, 1950, 0, 0, 0]),
                    ),
                ),
                np.asarray([3704, 0, 0, 0, 1950, 0, 0, 0]),
            )
            expense_array_without_dup = expense_array[:, [0, 4, 5, 6, 7]]
            sum_expenses = np.sum(expense_array_without_dup, axis=1)
            total_income = np.sum(income_array, axis=1)
            sanity_check_array = np.greater_equal(total_income, sum_expenses)
            temp_df["Unimportant"] = sanity_check_array.tolist()
            reduction = temp_df[temp_df.Unimportant]
            drop_reduction = reduction.drop("Unimportant", axis=1)

            # check that the food expense is larger than all the sub expenses
            expense_reduced_array = drop_reduction["Output"].to_numpy()
            expense_reduced_array = np.add(
                np.multiply(
                    expense_reduced_array,
                    np.subtract(
                        np.asarray(
                            [
                                791848,
                                437467,
                                140992,
                                74800,
                                2188560,
                                1049275,
                                149940,
                                731000,
                            ]
                        ),
                        np.asarray([3704, 0, 0, 0, 1950, 0, 0, 0]),
                    ),
                ),
                np.asarray([3704, 0, 0, 0, 1950, 0, 0, 0]),
            )
            food_mul_expense_array = expense_reduced_array[:, [1, 2, 3]]
            food_mul_expense_array_sum = np.sum(food_mul_expense_array, axis=1)
            food_expense_array = expense_reduced_array[:, 0]
            sanity_check_array = np.greater_equal(
                food_expense_array, food_mul_expense_array_sum
            )
            drop_reduction["Unimportant"] = sanity_check_array.tolist()
            new_reduction = drop_reduction[drop_reduction.Unimportant]
            satisfied_constraints_df = new_reduction.drop("Unimportant", axis=1)

            return satisfied_constraints_df

        def sample_2500_examples(df: pd.DataFrame) -> pd.DataFrame:
            """Sample 2500 examples from the dataframe without replacement."""
            temp_df = df.copy()
            sample_df = temp_df.sample(n=2500, replace=False, random_state=3, axis=0)
            return sample_df

        def add_input_output_family_income(df: pd.DataFrame) -> pd.DataFrame:
            """Add a multiindex denoting if the column is an input or output variable."""
            # copy the dataframe
            temp_df = df.copy()
            # extract all the column names
            column_names = temp_df.columns.tolist()
            # the 2nd-9th columns correspond to output variables and all others to input variables. So make list of corresponding lengths of 'Input' and 'Output'
            input_list_start = ["Input"]
            input_list_end = ["Input"] * (len(column_names) - 9)
            output_list = ["Output"] * 8
            # concat both lists
            input_output_list = input_list_start + output_list + input_list_end
            # define multi index for attaching this 'Input' and 'Output' list with the column names already existing
            multiindex_bias = pd.MultiIndex.from_arrays(
                [input_output_list, column_names]
            )
            # transpose such that index can be adjusted to multi index
            new_df = pd.DataFrame(df.transpose().to_numpy(), index=multiindex_bias)
            # transpose back such that columns are the same as before except with different labels
            return new_df.transpose()

        return (
            # drop missing values
            df.dropna(how="any")
            # convert object to fitting dtype
            .convert_dtypes()
            # remove all strings (no other dtypes are present except for integers and floats)
            .select_dtypes(exclude=["string"])
            # transform all numbers to same dtype
            .astype("float32")
            # drop column with label Agricultural Household indicator because this is not really a numerical input but rather a categorical/classification
            .drop(["Agricultural Household indicator"], axis=1, inplace=False)
            # this column is dropped because it depends on Agricultural Household indicator
            .drop(["Crop Farming and Gardening expenses"], axis=1, inplace=False)
            # use 8 output variables and 24 input variables
            .drop(
                [
                    "Total Rice Expenditure",
                    "Total Fish and  marine products Expenditure",
                    "Fruit Expenditure",
                    "Restaurant and hotels Expenditure",
                    "Alcoholic Beverages Expenditure",
                    "Tobacco Expenditure",
                    "Clothing, Footwear and Other Wear Expenditure",
                    "Imputed House Rental Value",
                    "Transportation Expenditure",
                    "Miscellaneous Goods and Services Expenditure",
                    "Special Occasions Expenditure",
                ],
                axis=1,
                inplace=False,
            )
            # add input and output labels to each column
            .pipe(add_input_output_family_income)
            # normalize all the columns
            .pipe(normalize_columns_income)
            # remove all datapoints that do not satisfy the constraints
            .pipe(check_constraints_income)
            # sample 2500 examples
            .pipe(sample_2500_examples)
        )



from abc import ABC, abstractmethod
from numbers import Number
import random
import string
from typing import Callable, Dict
from torch import Tensor, ge, gt, lt, le, zeros, FloatTensor, ones, tensor, float32
import logging
from torch.nn.functional import normalize

from .descriptor import Descriptor


class Constraint(ABC):
    """
    Abstract base class for defining constraints that can be applied during optimization in the constraint-guided gradient descent process.

    A constraint guides the optimization by evaluating the model's predictions and adjusting the loss based on certain conditions.
    Constraints can be applied to specific layers or neurons of the model, and they are scaled by a rescale factor to control the influence of the constraint on the overall loss.

    Attributes:
        descriptor (Descriptor): The descriptor object that provides a mapping of neurons to layers.
        constraint_name (str): A unique name for the constraint, which can be provided or generated automatically.
        rescale_factor (float): A factor used to scale the influence of the constraint on the overall loss.
        neuron_names (set[str]): A set of neuron names that are involved in the constraint.
        layers (set): A set of layers associated with the neurons specified in `neuron_names`.
    """

    descriptor: Descriptor = None

    def __init__(
        self,
        neuron_names: set[str],
        constraint_name: str = None,
        rescale_factor: float = 1.5,
    ) -> None:
        """
        Initializes the Constraint object with the given neuron names, constraint name, and rescale factor.

        Args:
            neuron_names (set[str]): A set of neuron names that are affected by the constraint.
            constraint_name (str, optional): A custom name for the constraint. If not provided, a random name is generated.
            rescale_factor (float, optional): A factor that scales the influence of the constraint. Defaults to 1.5.

        Raises:
            ValueError: If the descriptor has not been set or if a neuron name is not found in the descriptor.
        """

        # Init parent class
        super().__init__()

        # Init object variables
        self.rescale_factor = rescale_factor
        self.neuron_names = neuron_names

        # Perform checks
        if rescale_factor <= 1:
            logging.warning(
                f"Rescale factor for constraint {constraint_name} is <= 1. The network will favour general loss over the constraint-adjusted loss. Is this intended behaviour? Normally, the loss should always be larger than 1."
            )

        # If no constraint_name is set, generate one based on the class name and a random suffix
        if constraint_name:
            self.constraint_name = constraint_name
        else:
            random_suffix = "".join(
                random.choices(string.ascii_uppercase + string.digits, k=6)
            )
            self.constraint_name = f"{self.__class__.__name__}_{random_suffix}"
            logging.warning(
                f"Name for constraint is not set. Using {self.constraint_name}."
            )

        if self.descriptor == None:
            raise ValueError(
                "The descriptor of the base Constraint class in not set. Please assign the descriptor to the general Constraint class with 'Constraint.descriptor = descriptor' before defining network-specific contraints."
            )

        if not rescale_factor > 1:
            self.rescale_factor = abs(rescale_factor) + 1.5
            logging.warning(
                f"Rescale factor for constraint {constraint_name} is < 1, adjusted value {rescale_factor} to {self.rescale_factor}."
            )
        else:
            self.rescale_factor = rescale_factor

        self.neuron_names = neuron_names

        self.run_init_descriptor()

    def run_init_descriptor(self) -> None:
        """
        Initializes the layers associated with the constraint by mapping the neuron names to their corresponding layers
        from the descriptor.

        This method populates the `layers` attribute with layers associated with the neuron names provided in the constraint.

        Raises:
            ValueError: If a neuron name is not found in the descriptor's mapping of neurons to layers.
        """

        self.layers = set()
        for neuron_name in self.neuron_names:
            if neuron_name in self.descriptor.neuron_to_layer.keys():
                self.layers.add(self.descriptor.neuron_to_layer[neuron_name])
            else:
                raise ValueError(
                    f'The neuron name {neuron_name} used with constraint {self.constraint_name} is not defined in the descriptor. Please add it to the correct layer using descriptor.add("layer", ...).'
                )

    @abstractmethod
    def check_constraint(self, prediction: dict[str, Tensor]) -> Dict[str, Tensor]:
        """
        Abstract method to check if the constraint is satisfied based on the model's predictions.

        This method should be implemented in subclasses to define the specific logic for evaluating the constraint based on the model's predictions.

        Args:
            prediction (dict[str, Tensor]): A dictionary of model predictions, indexed by layer names.

        Returns:
            dict[str, Tensor]: A dictionary containing the satisfaction status of the constraint for each layer or neuron.

        Raises:
            NotImplementedError: If the method is not implemented in a subclass.
        """

        raise NotImplementedError

    @abstractmethod
    def calculate_direction(self, prediction: dict[str, Tensor]) -> Dict[str, Tensor]:
        """
        Abstract method to calculate the direction in which the model's predictions need to be adjusted to satisfy the constraint.

        This method should be implemented in subclasses to define how to adjust the model's predictions based on the constraint.

        Args:
            prediction (dict[str, Tensor]): A dictionary of model predictions, indexed by layer names.

        Returns:
            dict[str, Tensor]: A dictionary containing the direction for each layer or neuron, to adjust the model's predictions.

        Raises:
            NotImplementedError: If the method is not implemented in a subclass.
        """
        raise NotImplementedError


class ScalarConstraint(Constraint):
    """
    A subclass of the `Constraint` class that applies a scalar constraint on a specific neuron in the model.

    This constraint compares the value of a specific neuron in the model to a scalar value using a specified comparator (e.g., greater than, less than).
    If the constraint is violated, it adjusts the loss according to the direction defined by the comparator.

    Attributes:
        comparator (Callable[[Tensor, Number], Tensor]): A comparator function (e.g., greater than, less than) to evaluate the constraint.
        scalar (Number): The scalar value to compare the neuron value against.
        direction (int): The direction in which the constraint should adjust the model's predictions (either 1 or -1 based on the comparator).
        layer (str): The layer associated with the specified neuron.
        index (int): The index of the specified neuron within the layer.
    """

    def __init__(
        self,
        neuron_name: str,
        comparator: Callable[[Tensor, Number], Tensor],
        scalar: Number,
        name: str = None,
        descriptor: Descriptor = None,
        rescale_factor: float = 1.5,
    ) -> None:
        """
        Initializes the ScalarConstraint with the given neuron name, comparator, scalar value, and other optional parameters.

        Args:
            neuron_name (str): The name of the neuron that the constraint applies to.
            comparator (Callable[[Tensor, Number], Tensor]): The comparator function used to evaluate the constraint (e.g., ge, le, gt, lt).
            scalar (Number): The scalar value that the neuron value is compared to.
            name (str, optional): A custom name for the constraint. If not provided, a name is generated based on the neuron name, comparator, and scalar.
            descriptor (Descriptor, optional): The descriptor that maps neurons to layers. If not provided, the global descriptor is used.
            rescale_factor (float, optional): A factor that scales the influence of the constraint on the overall loss. Defaults to 1.5.

        Raises:
            ValueError: If the comparator function is not one of the supported comparison operators (ge, le, gt, lt).
        """

        # Compose constraint name
        name = f"{neuron_name}_{comparator.__name__}_{str(scalar)}"

        # Init parent class
        super().__init__({neuron_name}, name, rescale_factor)

        # Init variables
        self.comparator = comparator
        self.scalar = scalar

        if descriptor != None:
            self.descriptor = descriptor
            self.run_init_descriptor()

        # Get layer name and feature index from neuron_name
        self.layer = self.descriptor.neuron_to_layer[neuron_name]
        self.index = self.descriptor.neuron_to_index[neuron_name]

        # If comparator function is not supported, raise error
        if comparator not in [ge, le, gt, lt]:
            raise ValueError(
                f"Comparator {str(comparator)} used for constraint {name} is not supported. Only ge, le, gt, lt are allowed."
            )

        # Calculate directions based on constraint operator
        if self.comparator in [lt, le]:
            self.direction = 1
        elif self.comparator in [gt, ge]:
            self.direction = -1

    def check_constraint(self, prediction: dict[str, Tensor]) -> Dict[str, Tensor]:
        """
        Checks if the constraint is satisfied based on the model's predictions.

        The constraint is evaluated by applying the comparator to the value of the specified neuron and the scalar value.

        Args:
            prediction (dict[str, Tensor]): A dictionary of model predictions, indexed by layer names.

        Returns:
            dict[str, Tensor]: A dictionary containing the constraint satisfaction result for the specified layer.
        """

        result = ~self.comparator(prediction[self.layer][:, self.index], self.scalar)

        return {self.layer: result}

    def calculate_direction(self, prediction: dict[str, Tensor]) -> Dict[str, Tensor]:
        """
        Calculates the direction in which the model's predictions need to be adjusted to satisfy the constraint.

        The direction is determined by the comparator and represents either a positive or negative adjustment.

        Args:
            prediction (dict[str, Tensor]): A dictionary of model predictions, indexed by layer names.

        Returns:
            dict[str, Tensor]: A dictionary containing the direction for each layer or neuron, to adjust the model's predictions.
        """

        output = zeros(
            prediction[self.layer].size(),
            device=prediction[self.layer].device,
        )
        output[:, self.index] = self.direction

        return {self.layer: output}


class BinaryConstraint(Constraint):
    """
    A class representing a binary constraint between two neurons in a neural network.

    This class checks and enforces a constraint between two neurons using a
    comparator function. The constraint is applied between two neurons located
    in different layers of the neural network. The class also calculates the
    direction for gradient adjustment based on the comparator.

    Attributes:
        neuron_name_left (str): The name of the first neuron involved in the constraint.
        neuron_name_right (str): The name of the second neuron involved in the constraint.
        comparator (Callable[[Tensor, Number], Tensor]): A function that compares the values of the two neurons.
        layer_left (str): The layer name for the first neuron.
        layer_right (str): The layer name for the second neuron.
        index_left (int): The index of the first neuron within its layer.
        index_right (int): The index of the second neuron within its layer.
        direction_left (float): The normalized direction for gradient adjustment of the first neuron.
        direction_right (float): The normalized direction for gradient adjustment of the second neuron.
    """

    def __init__(
        self,
        neuron_name_left: str,
        comparator: Callable[[Tensor, Number], Tensor],
        neuron_name_right: str,
        name: str = None,
        descriptor: Descriptor = None,
        rescale_factor: float = 1.5,
    ) -> None:
        """
        Initializes the binary constraint with two neurons, a comparator, and other configuration options.

        Args:
            neuron_name_left (str): The name of the first neuron in the constraint.
            comparator (Callable[[Tensor, Number], Tensor]): A function that compares the values of the two neurons.
            neuron_name_right (str): The name of the second neuron in the constraint.
            name (str, optional): The name of the constraint. If not provided, a default name is generated.
            descriptor (Descriptor, optional): The descriptor containing the mapping of neurons to layers.
            rescale_factor (float, optional): A factor to rescale the constraint value. Default is 1.5.
        """

        # Compose constraint name
        name = f"{neuron_name_left}_{comparator.__name__}_{neuron_name_right}"

        # Init parent class
        super().__init__({neuron_name_left, neuron_name_right}, name, rescale_factor)

        # Init variables
        self.comparator = comparator

        if descriptor != None:
            self.descriptor = descriptor
            self.run_init_descriptor()

        # Get layer name and feature index from neuron_name
        self.layer_left = self.descriptor.neuron_to_layer[neuron_name_left]
        self.layer_right = self.descriptor.neuron_to_layer[neuron_name_right]
        self.index_left = self.descriptor.neuron_to_index[neuron_name_left]
        self.index_right = self.descriptor.neuron_to_index[neuron_name_right]

        # If comparator function is not supported, raise error
        if comparator not in [ge, le, gt, lt]:
            raise RuntimeError(
                f"Comparator {str(comparator)} used for constraint {name} is not supported. Only ge, le, gt, lt are allowed."
            )

        # Calculate directions based on constraint operator
        if self.comparator in [lt, le]:
            self.direction_left = -1
            self.direction_right = 1
        else:
            self.direction_left = 1
            self.direction_right = -1

        # Normalize directions
        normalized_directions = normalize(
            tensor([self.direction_left, self.direction_right]).type(float32),
            p=2,
            dim=0,
        )
        self.direction_left = normalized_directions[0]
        self.direction_right = normalized_directions[1]

    def check_constraint(self, prediction: dict[str, Tensor]) -> Dict[str, Tensor]:
        """
        Checks whether the binary constraint is satisfied between the two neurons.

        This function applies the comparator to the output values of the two neurons
        and returns a Boolean result for each neuron.

        Args:
            prediction (dict[str, Tensor]): A dictionary containing the predictions for each layer.

        Returns:
            dict[str, Tensor]: A dictionary with the layer names as keys and the constraint satisfaction results as values.
        """

        result = ~self.comparator(
            prediction[self.layer_left][:, self.index_left],
            prediction[self.layer_right][:, self.index_right],
        )

        return {self.layer_left: result, self.layer_right: result}

    def calculate_direction(self, prediction: dict[str, Tensor]) -> Dict[str, Tensor]:
        """
        Calculates the direction for gradient adjustment for both neurons involved in the constraint.

        The directions are normalized and represent the direction in which the constraint should be enforced.

        Args:
            prediction (dict[str, Tensor]): A dictionary containing the predictions for each layer.

        Returns:
            dict[str, Tensor]: A dictionary with the layer names as keys and the direction vectors as values.
        """

        output_left = zeros(
            prediction[self.layer_left].size(),
            device=prediction[self.layer_left].device,
        )
        output_left[:, self.index_left] = self.direction_left

        output_right = zeros(
            prediction[self.layer_right].size(),
            device=prediction[self.layer_right].device,
        )
        output_right[:, self.index_right] = self.direction_right

        return {self.layer_left: output_left, self.layer_right: output_right}


# FIXME
class SumConstraint(Constraint):
    def __init__(
        self,
        neuron_names_left: list[str],
        comparator: Callable[[Tensor, Number], Tensor],
        neuron_names_right: list[str],
        weights_left: list[float] = None,
        weights_right: list[float] = None,
        name: str = None,
        descriptor: Descriptor = None,
        rescale_factor: float = 1.5,
    ) -> None:

        # Init parent class
        super().__init__(
            set(neuron_names_left) & set(neuron_names_right), name, rescale_factor
        )

        # Init variables
        self.comparator = comparator

        if descriptor != None:
            self.descriptor = descriptor
            self.run_init_descriptor()

        # Get layer names and feature indices from neuron_name
        self.layers_left = []
        self.indices_left = []
        for neuron_name in neuron_names_left:
            self.layers_left.append(self.descriptor.neuron_to_layer[neuron_name])
            self.indices_left.append(self.descriptor.neuron_to_index[neuron_name])

        self.layers_right = []
        self.indices_right = []
        for neuron_name in neuron_names_right:
            self.layers_right.append(self.descriptor.neuron_to_layer[neuron_name])
            self.indices_right.append(self.descriptor.neuron_to_index[neuron_name])

        # If comparator function is not supported, raise error
        if comparator not in [ge, le, gt, lt]:
            raise ValueError(
                f"Comparator {str(comparator)} used for constraint {name} is not supported. Only ge, le, gt, lt are allowed."
            )

        # If feature list dimensions don't match weight list dimensions, raise error
        if weights_left and (len(neuron_names_left) != len(weights_left)):
            raise ValueError(
                "The dimensions of neuron_names_left don't match with the dimensions of weights_left."
            )
        if weights_right and (len(neuron_names_right) != len(weights_right)):
            raise ValueError(
                "The dimensions of neuron_names_right don't match with the dimensions of weights_right."
            )

        # If weights are provided for summation, transform them to Tensors
        if weights_left:
            self.weights_left = FloatTensor(weights_left)
        else:
            self.weights_left = ones(len(neuron_names_left))
        if weights_right:
            self.weights_right = FloatTensor(weights_right)
        else:
            self.weights_right = ones(len(neuron_names_right))

        # Calculate directions based on constraint operator
        if self.comparator in [lt, le]:
            self.direction_left = -1
            self.direction_right = 1
        else:
            self.direction_left = 1
            self.direction_right = -1

        # Normalize directions
        normalized_directions = normalize(
            tensor(self.direction_left, self.direction_right), p=2, dim=0
        )
        self.direction_left = normalized_directions[0]
        self.direction_right = normalized_directions[1]

    def check_constraint(self, prediction: dict[str, Tensor]) -> Dict[str, Tensor]:
        raise NotImplementedError
        # # TODO remove the dynamic to device conversion and do this in initialization one way or another
        # weighted_sum_left = (
        #     prediction[layer_left][:, index_left]
        #     * self.weights_left.to(prediction[layer_left].device)
        # ).sum(dim=1)
        # weighted_sum_right = (
        #     prediction[layer_right][:, index_right]
        #     * self.weights_right.to(prediction[layer_right].device)
        # ).sum(dim=1)

        # result = ~self.comparator(weighted_sum_left, weighted_sum_right)

        # return {layer_left: result, layer_right: result}
        pass

    def calculate_direction(self, prediction: dict[str, Tensor]) -> Dict[str, Tensor]:
        raise NotImplementedError
        # # TODO move this to constructor somehow
        # layer_left = prediction.neuron_to_layer[self.neuron_name_left]
        # layer_right = prediction.neuron_to_layer[self.neuron_name_right]
        # index_left = prediction.neuron_to_index[self.neuron_name_left]
        # index_right = prediction.neuron_to_index[self.neuron_name_right]

        # output_left = zeros(
        #     prediction[layer_left].size(),
        #     device=prediction[layer_left].device,
        # )
        # output_left[:, index_left] = self.direction_left

        # output_right = zeros(
        #     prediction.layer_to_data[layer_right].size(),
        #     device=prediction.layer_to_data[layer_right].device,
        # )
        # output_right[:, index_right] = self.direction_right

        # return {layer_left: output_left, layer_right: output_right}
        pass



import logging
from typing import Union
from torch import Tensor
from torch.nn import Module
from torch.nn.modules.loss import _Loss
from torch.optim import Optimizer

from .core import CGGDModule
from .constraints import Constraint
from .descriptor import Descriptor


class Learner(CGGDModule):
    def __init__(
        self,
        network: Module,
        descriptor: Descriptor,
        constraints: list[Constraint],
        loss_function: Union[_Loss, dict[str, _Loss]],
        optimizer: Optimizer,
    ):
        """
        A class that integrates a neural network with a training and validation loop,
        supporting single or multi-output loss functions. The class manages the forward pass,
        training step, and validation step while also configuring the optimizer.

        Args:
            network (Module): The neural network model to be trained.
            descriptor (Descriptor): An object that defines the structure of the network,
                                    including the output layers.
            constraints (list[Constraint]): A list of constraints that can be applied during training.
            loss_function (Union[_Loss, dict[str, _Loss]]): A loss function or a dictionary of loss functions
                                                        for each output layer.
            optimizer (Optimizer): The optimizer used for training the model.

        Raises:
            ValueError: If the descriptor does not contain any output layers or if the number of loss functions
                        does not match the number of output layers when using a dictionary of loss functions.
        """

        # Init parent class
        super().__init__(descriptor, constraints)

        # Init object variables
        self.network = network
        self.descriptor = descriptor
        self.loss_function = loss_function
        self.optimizer = optimizer

        # Perform checks
        if len(self.descriptor.output_layers) == 0:
            raise ValueError(
                'The descriptor class must contain one or more output layers. Mark a layer as output by setting descriptor.add("layer", ..., output=True).'
            )

        if isinstance(loss_function, _Loss):
            if len(self.descriptor.output_layers) > 1:
                logging.warning(
                    f"Multiple layers were marked as output, but only one loss function is defined. Only the loss of layer {list(self.descriptor.output_layers)[0]} will be calculated and used. To use the same loss function for all output layers, please specify then explicitly."
                )

        if isinstance(loss_function, dict):
            if len(self.descriptor.output_layers) != len(loss_function):
                raise ValueError(
                    f"The number of marked output layers does not match the number of provided loss functions."
                )

        # Assign proper step function based on if one or multiple loss functions are assigned
        if isinstance(loss_function, _Loss):
            self.training_step = self.training_step_single
            self.validation_step = self.validation_step_single

        if isinstance(loss_function, dict):
            self.training_step = self.training_step_multi
            self.validation_step = self.validation_step_multi

    def forward(self, x):
        """
        Perform a forward pass through the network.

        Args:
            x (Tensor): The input tensor to pass through the network.

        Returns:
            Tensor: The model's output for the given input.
        """

        return self.network(x)

    def training_step_single(self, batch, batch_idx):
        """
        Perform a single training step using a single loss function.

        Args:
            batch (tuple): A tuple containing the input and target output tensors.
            batch_idx (int): The index of the batch in the current epoch.

        Returns:
            Tensor: The loss value for the batch.
        """

        self.train()

        inputs, outputs = batch
        prediction: dict[str, Tensor] = self(inputs)

        layer = list(self.descriptor.output_layers)[0]
        loss = self.loss_function(prediction[layer], outputs)

        self.log(
            "train_loss",
            loss,
            on_step=False,
            on_epoch=True,
        )

        return super().training_step(prediction, loss)

    def training_step_multi(self, batch, batch_idx):
        """
        Perform a training step using multiple loss functions, one for each output layer.

        Args:
            batch (tuple): A tuple containing the input and target output tensors.
            batch_idx (int): The index of the batch in the current epoch.

        Returns:
            Tensor: The total loss value for the batch, combining the losses from all output layers.
        """

        self.train()

        inputs, outputs = batch
        prediction: dict[str, Tensor] = self(inputs)

        # TODO add hyperparameter to scale loss per function
        loss = 0
        for layer in self.descriptor.output_layers:
            layer_loss = self.loss_function[layer](prediction[layer], outputs)
            loss += layer_loss

            self.log(
                f"train_loss_{layer}",
                layer_loss,
                on_step=False,
                on_epoch=True,
            )

        self.log(
            "train_loss",
            loss,
            on_step=False,
            on_epoch=True,
        )

        return super().training_step(prediction, loss)

    def validation_step_single(self, batch, batch_idx):
        """
        Perform a single validation step using a single loss function.

        Args:
            batch (tuple): A tuple containing the input and target output tensors.
            batch_idx (int): The index of the batch in the current epoch.

        Returns:
            Tensor: The validation loss for the batch.
        """

        self.eval()

        inputs, outputs = batch
        prediction: dict[str, Tensor] = self(inputs)

        layer = list(self.descriptor.output_layers)[0]
        loss = self.loss_function(prediction[layer], outputs)

        self.log(
            "valid_loss",
            loss,
            on_step=False,
            on_epoch=True,
        )

        return super().validation_step(prediction, loss)

    def validation_step_multi(self, batch, batch_idx):
        """
        Perform a validation step using multiple loss functions, one for each output layer.

        Args:
            batch (tuple): A tuple containing the input and target output tensors.
            batch_idx (int): The index of the batch in the current epoch.

        Returns:
            Tensor: The total validation loss for the batch, combining the losses from all output layers.
        """

        self.eval()

        inputs, outputs = batch
        prediction: dict[str, Tensor] = self(inputs)

        loss = 0
        for layer in self.descriptor.output_layers:
            layer_loss = self.loss_function[layer](prediction[layer], outputs)
            loss += layer_loss

            self.log(
                f"valid_loss_{layer}",
                layer_loss,
                on_step=False,
                on_epoch=True,
            )

        self.log(
            "valid_loss",
            loss,
            on_step=False,
            on_epoch=True,
        )

        return super().validation_step(prediction, loss)

    def configure_optimizers(self):
        """
        Configure the optimizer for training.

        Returns:
            Optimizer: The optimizer used to update the model's parameters during training.
        """

        return self.optimizer



from torch import Tensor, tensor, sum, numel
from torchmetrics import Metric

# NOTE


class ConstraintSatisfactionRatio(Metric):
    """
    A custom metric to calculate the ratio of satisfied constraints in a neural network model.
    It computes the proportion of constraints that have been satisfied,
    where satisfaction is determined based on the provided constraint results.

    This metric tracks the number of unsatisfied constraints and the total number of constraints
    during the training process, and computes the ratio of satisfied constraints once all updates
    have been made.

    Attributes:
        unsatisfied (Tensor): Tracks the number of unsatisfied constraints.
        total (Tensor): Tracks the total number of constraints processed.

    Note:
        For more information about custom metrics, we refer to the Pytorch Lightning documentation
        at https://lightning.ai/docs/torchmetrics/stable/pages/implement.html
    """

    def __init__(self, **kwargs):
        """
        Initializes the ConstraintSatisfactionRatio metric by setting up the
        state variables to track the number of unsatisfied and total constraints.

        Args:
            **kwargs: Additional arguments to pass to the base Metric class constructor.
        """

        # Init parent class
        super().__init__(**kwargs)

        # Init scalar tensors that will hold metric values
        self.add_state("unsatisfied", default=tensor(0), dist_reduce_fx="sum")
        self.add_state("total", default=tensor(0), dist_reduce_fx="sum")

    def update(self, constraint_result: Tensor) -> None:
        """
        Updates the state of the metric with the latest constraint results.

        Args:
            constraint_result (Tensor): A tensor representing the result of
                                         the constraint checks, where each
                                         element indicates whether a constraint
                                         is satisfied (e.g., 0 for satisfied,
                                         1 for unsatisfied).
        """
        self.unsatisfied += sum(constraint_result)
        self.total += numel(constraint_result)

    def compute(self) -> Tensor:
        """
        Computes the constraint satisfaction ratio, defined as:
        1 - (number of unsatisfied constraints / total constraints).

        Returns:
            Tensor: The satisfaction ratio as a scalar tensor.
        """
        return 1 - (self.unsatisfied.float() / self.total)
